/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.api;

/**
 * This interface is used for QueryTrees which can be optimized.
 */
public interface EvaluationMetaData {

	/**
	 * Set a boolean for declaring the use of a cache for intermediate
	 * materialization of results at evaluation time.
	 */
	void setCache(boolean isCache);

	/**
	 * Get the value of the boolean indicating whether cache is used for
	 * materialization of intermediate results at evaluation time.
	 * @return the boolean indicating whether the cache is used for
	 * materialization of intermediate results.
	 */
	boolean isCache();

	/**
	 * Set a boolean for declaring parallel evaluation of sub-QueryTrees.
	 */
	void setParallel(boolean isTupleLoader);

	/**
	 * Get the value of the boolean indicating whether parallel evaluation
	 * of sub-QueryTrees is performed.
	 * @return the boolean indicating parallel evaluation of sub-QueryTrees is
	 * performed.
	 */
	boolean isParallel();

	/**
	 * Gets the value of the boolean indicating whether a limited range of
	 * evaluation is performed.
	 * <p>If true, there should be a value for the first tuple index and a
	 * value for the maximum number of tuples to be returned by
	 * the evaluation.</p>
	 * @return true if limited range evaluation is to be performed.
	 */
	boolean isLimitedRange();

	/**
	 * Sets the value for the rank of the first tuple to be returned by the
	 * evaluation in the case of a range-limited evaluation.
	 * @param start the index of the first tuple to be returned by the
	 * evaluation, stating at 0.
	 */
	void setLimitedRangeStartAt(int start);

	/**
	 * Gets the value for the rank of the first tuple to be returned by the
	 * evaluation in the case of a range-limited evaluation.
	 * @return the value for the first tuple to be returned by the evaluation,
	 * starting at 0.
	 */
	int getLimitedRangeStartAt();

	/**
	 * Sets the maximum number of tuples to be returned by the evaluation in
	 * the case of a range-limited evaluation.
	 * @param number the maximum number of tuples to be returned by the
	 * evaluation
	 */
	void setLimitedRangeSize(int number);

	/**
	 * Gets the maximum number of tuples to be returned by the evaluation in
	 * the case of a range-limited evaluation.
	 * @return the maximum number of tuples to be returned by the evaluation.
	 */
	int getLimitedRangeSize();

}
