/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import java.util.ArrayList;

import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.EncodePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.Avg;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.Count;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.Sum;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.lib.IndexesGenerator;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.lib.Nest;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.query.rdb.lib.AggregateRdbQueryNode;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class TestJormQueryRewriter extends TestJormPDHelper {

	public TestJormQueryRewriter(String testName, String loggerName) {
		super(testName, loggerName);
	}

	public abstract QueryLeaf getQueryLeafA(Example e) throws MedorException;

	public abstract QueryLeaf getQueryLeafAgg(Example e) throws MedorException;

	public abstract QueryLeaf getQueryLeafAggParam(Example e) throws MedorException;

	public abstract QueryLeaf getQueryLeafAggDirect(Example e)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafAggWithWhere(Example e)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafOrderByA(Example e)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafMemberOfBD(Example eB, Example eD)
			throws Exception;

	public abstract QueryLeaf getQueryLeafMemberOfBC(Example eB, Example eC)
			throws Exception;

	public abstract QueryLeaf getQueryLeafMOfParamBD(Example eB, Example eD)
			throws Exception;

	public abstract QueryLeaf getQueryLeafAnimal(Example e)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafLongLong(Example e)
			throws MedorException;


	public void test1() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "test1: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafA(e);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void test1distinct() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "");
			qn.setDistinct(true);
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "test1distinct: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafA(e);
			QueryNode expected = new SelectProject("");
			expected.setDistinct(true);
			leaf.setDistinct(true);
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1distinct: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testOrderBy() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryTreeField field1 =
					(QueryTreeField)e.extent.getField(e.clazz.getFQName() + ".field1");
			QueryTreeField field2 =
					(QueryTreeField)e.extent.getField(e.clazz.getFQName() + ".field2");
			// Add an OrderBy Field
			OrderField of1 = new BasicOrderField(field1, true);
			OrderField of2 = new BasicOrderField(field2);
			qn.setOrderBy(new OrderField[] {of1, of2});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "test1distinct: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafOrderByA(e);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1distinct: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	//TODO corrected expected query tree
	public void _testOrderByWithNode() {
		try {
			Example eA = getExtentOfA();
			Example eB = getExtentOfB();
			Example eC = getExtentOfC();
			Expression exp2 = new Equal(
					new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".y")),
					new BasicFieldOperand(eC.extent.getField(eC.clazz.getFQName() + ".x")));
			QueryNode qn2 = new SelectProject(exp2, "");
			QueryTreeField qtf2 = (QueryTreeField)
					eB.extent.getField(eB.extent.getPNameFieldName());
			QueryTreeField fqn2 = qn2.addPropagatedField(
					qtf2.getName(),
					qtf2.getType(),
					new QueryTreeField[]{qtf2});
			Expression exp = new Equal(
					new BasicFieldOperand(eA.extent.getField(eA.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					eA.extent.getField(eA.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			qn.addPropagatedField(
					fqn2.getName(),
					fqn2.getType(),
					new QueryTreeField[]{fqn2});
			QueryTreeField field1 =
					(QueryTreeField)eA.extent.getField(eA.clazz.getFQName() + ".field1");
			QueryTreeField field2 =
					(QueryTreeField)eA.extent.getField(eA.clazz.getFQName() + ".field2");
			// Add an OrderBy Field
			OrderField of1 = new BasicOrderField(field1, true);
			OrderField of2 = new BasicOrderField(field2);
			qn.setOrderBy(new OrderField[] {of1, of2});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "test1distinct: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eA.extent.setPMapper(mapper, PROJECT_NAME);
			eA.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafOrderByA(eA);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					eA.clazz.getFQName() + ".object",
					eA.extent.getField(eA.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(eA.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							eA.extent.getField(eA.clazz.getFQName() + ".object").getType()
					)
			);

			logger.log(BasicLevel.DEBUG, "test1distinct: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}
	public void testMemberOfsingleBC() throws Exception {
		try {
			Example eB = getExtentOfB(new String[] {"y"});
			Example eC = getExtentOfC();
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".y")));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eC.extent.getField(eC.clazz.getFQName() + ".x")));
			Expression exp1 = new MemberOf(al1, al2);
			Expression exp = new And(exp1,
					new Equal(new BasicFieldOperand(eC.extent.getField(eC.clazz.getFQName() + ".x")),
							new BasicOperand((long) 2)));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					eB.extent.getField(eB.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testMemberOfSingleBC: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName(PROJECT_NAME);
			eC.extent.setPMapper(mapper, PROJECT_NAME);
			eC.extent.setProjectName(PROJECT_NAME);
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafMemberOfBC(eB, eC);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					eB.clazz.getFQName() + ".object",
					eB.extent.getField(eB.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eB.extent.getField(eB.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOfSingleBC: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testIsNullOnPName() throws Exception {
		try {
			Example eLL = getExtentOfLongLong();
			eLL.extent.setPMapper(mapper, PROJECT_NAME);
			eLL.extent.setProjectName(PROJECT_NAME);
			QueryTreeField qtf = (QueryTreeField)
					eLL.extent.getField(eLL.extent.getPNameFieldName());
			FieldOperand bPName = new BasicFieldOperand(qtf);

			PolymorphIdBinder prnc = new PolymorphIdBinder();
			PName pnamenull = prnc.getNull();

			Expression exp = new Equal(bPName,
					new BasicOperand(pnamenull,	PTypeSpaceMedor.PNAME));
			QueryNode qn = new SelectProject(exp, "");
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testIsNullOnPName: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eLL.extent.setPMapper(mapper, PROJECT_NAME);
			eLL.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//building the expected result
			QueryNode expected = new SelectProject("");
			QueryLeaf leaf = getQueryLeafLongLong(eLL);

			expected.addCalculatedField(
					eLL.clazz.getFQName() + ".object",
					eLL.extent.getField(eLL.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eLL.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eLL.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eLL.extent.getField(eLL.clazz.getFQName() + ".object").getType()
					)
			);

			logger.log(BasicLevel.DEBUG, "testIsNullOnPName: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testMemberOfmultiBD() throws Exception {
		try {
			Example eB = getExtentOfB();
			Example eD = getExtentOfD();
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".object")));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eD.extent.getField(eD.clazz.getFQName() + ".object")));
			Expression exp = new MemberOf(al1, al2);
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					eB.extent.getField(eB.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testMemberOf: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName("test");
			eD.extent.setPMapper(mapper, PROJECT_NAME);
			eD.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//building the expected result
			QueryLeaf leaf = getQueryLeafMemberOfBD(eB, eD);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					eB.clazz.getFQName() + ".object",
					eB.extent.getField(eB.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eB.extent.getField(eB.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOf: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	/**
	 * TODO Handle MemberOf(param) in JormFlatten2Rdb for this test to pass
	 */
	public void _testMemberOfParamBD() throws Exception {
		try {
			logger.log(BasicLevel.DEBUG, "Entering testMemberOfParamBD");

			Example eB = getExtentOfB(new String[] {"y"});
			Example eD = getExtentOfD();
			ArrayList al1 = new ArrayList(1);
			BasicOperand value = new BasicOperand(1013);
			al1.add(value);
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".y")));
			Expression exp = new MemberOf(al1, al2);
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					eD.extent.getField(eD.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testMemberOfParamBD: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName("test");
			eD.extent.setPMapper(mapper, PROJECT_NAME);
			eD.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//building the expected result
			QueryLeaf leaf = getQueryLeafMOfParamBD(eB, eD);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					eD.clazz.getFQName() + ".object",
					eD.extent.getField(eD.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eD.clazz.getFQName()
										+ "." + "b")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eD.clazz.getFQName()
										+ "." + "a"))
							},
							new String[]{"b","a"},
							null,
							eD.extent.getField(eD.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOfParamBD: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	//TODO replace tables with JoinTables for the expected result
	public void _testAtoB() {
		try {
			Example a = getExtentOfAcolocB();
			Example b1 = getExtentOfBcolocC();
			Example b2 = getExtentOfBcolocC();
			Expression exp1 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b1")),
					new BasicFieldOperand(b1.extent.getField(b1.extent.getPNameFieldName()))
			);
			Expression exp2 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b2")),
					new BasicFieldOperand(b2.extent.getField(b2.extent.getPNameFieldName()))
			);
			Expression exp3 = new And(exp1, exp2);
			Expression exp4 = new Equal(
					new BasicFieldOperand(b1.extent.getField(b1.clazz.getFQName() + ".f1")),
					new BasicFieldOperand(b2.extent.getField(b2.clazz.getFQName() + ".f2"))
			);
			Expression exp = new And(exp3, exp4);
			QueryNode qn = new SelectProject(exp, "");

			QueryTreeField f = (QueryTreeField) a.extent.getField(a.extent.getPNameFieldName());
			qn.addPropagatedField(f.getName(), f.getType(), new QueryTreeField[]{f});

			QueryRewriter jqr = this.createQueryRewriter();
			/*
			logger.log(BasicLevel.DEBUG, "testAtoB: QueryTree before transformation");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(qn, logger);
			*/
			//assign mapper + project names to each extent
			a.extent.setPMapper(mapper, getMapperName());
			a.extent.setProjectName("test");
			b1.extent.setPMapper(mapper, getMapperName());
			b1.extent.setProjectName("test");
			b2.extent.setPMapper(mapper, getMapperName());
			b2.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);
			logger.log(BasicLevel.DEBUG, "=========================================");

			//TODO replace expected below with JoinedTables
			Example ea = getExtentOfAcolocB();
			QualifiedTable tableA = new BasicQualifiedTable("T_A", "T_A_0");
			String nameA = "org.objectweb.medor.optim.jorm.rdb.AcolocB";
			String prefixA = nameA + ".";
			QualifiedTable tableB1 = new BasicQualifiedTable("T_B", "T_B_1_b1_join");
			QualifiedTable tableB2 = new BasicQualifiedTable("T_B", "T_B_2_b2_join");
			String nameB = "org.objectweb.medor.optim.jorm.rdb.BcolocC";
			String prefixB = nameB + ".";
			RdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(null,
					new QualifiedTable[]{tableA, tableB1, tableB2}, "");
			RdbExpField idA = leaf.addRdbField(
					prefixA + "a_name", PTypeSpace.LONG, "A_PK", tableA);
			RdbExpField b1ToA = leaf.addRdbField(
					"BtoA_1_FK", PTypeSpace.LONG, "BtoA_1_FK", tableB1);
			RdbExpField f1 = leaf.addRdbField(
					prefixB + "f1", PTypeSpace.STRING, "col_f1", tableB1);
			leaf.removeRdbField(f1);
			RdbExpField b2ToA = leaf.addRdbField(
					"BtoA_2_FK", PTypeSpace.LONG, "BtoA_2_FK", tableB2);
			RdbExpField f2 = leaf.addRdbField(
					prefixB + "f2", PTypeSpace.STRING, "col_f2", tableB2);
			leaf.removeRdbField(f2);
			Expression eq3 = new Equal(
					new BasicFieldOperand(f1), new BasicFieldOperand(f2)
			);
			Expression eq1 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b1ToA)
			);
			Expression eq2 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b2ToA)
			);
			Expression expRdb = new And(new And(eq1, eq2), eq3);
			leaf.setQueryFilter(expRdb);
			QueryNode expected = new SelectProject("");
			SinglePName spn = new SinglePName(
					new BasicFieldOperand(idA),
					new BasicParameterOperand(
							PTypeSpaceMedor.NAMING_CONTEXT, prefixA + "object"),
					ea.extent.getField(prefixA + "object").getType());
			expected.addCalculatedField(prefixA + "object", spn.getType(), spn);

			logger.log(BasicLevel.DEBUG, "testAtoB: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
			logger.log(BasicLevel.DEBUG, "End of testAtoB");

		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void _testAtoBtoC() {
		try {
			Example a = getExtentOfAcolocB();
			Example b = getExtentOfBcolocC();
			Example c = getExtentOfCcoloc();
			Expression exp1 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b1")),
					new BasicFieldOperand(b.extent.getField(b.extent.getPNameFieldName()))
			);
			Expression exp2 = new Equal(
					new BasicFieldOperand(b.extent.getField(b.clazz.getFQName() + ".c")),
					new BasicFieldOperand(c.extent.getField(c.extent.getPNameFieldName()))
			);
			Expression exp = new And(exp1, exp2);
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField f = (QueryTreeField) a.extent.getField(a.extent.getPNameFieldName());
			QueryTreeField fc = (QueryTreeField) c.extent.getField(c.clazz.getFQName() + ".fc");
			qn.addPropagatedField(f.getName(), f.getType(), new QueryTreeField[]{f});
			qn.addPropagatedField(fc.getName(), fc.getType(), new QueryTreeField[]{fc});

			// Add an OrderBy Field
			OrderField of1 = new BasicOrderField((QueryTreeField)a.extent.getField(a.clazz.getFQName() + ".b1"), true);
			qn.setOrderBy(new OrderField[] {of1});

			QueryRewriter jqr = this.createQueryRewriter();
			logger.log(BasicLevel.DEBUG, "testBtoC: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			a.extent.setPMapper(mapper, getMapperName());
			a.extent.setProjectName("test");
			b.extent.setPMapper(mapper, getMapperName());
			b.extent.setProjectName("test");
			c.extent.setPMapper(mapper, getMapperName());
			c.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//TODO replace expected below with JoinedTables
			Example ea = getExtentOfAcolocB();
			QualifiedTable tableA = new BasicQualifiedTable("T_A", "T_A_0");
			String nameA = "org.objectweb.medor.optim.jorm.rdb.AcolocB";
			String prefixA = nameA + ".";
			QualifiedTable tableB = new BasicQualifiedTable("T_B", "T_B_1_b1_join");
			QualifiedTable tableC = new BasicQualifiedTable("T_C", "T_C_4_c_join");
			String nameB = "org.objectweb.medor.optim.jorm.rdb.BcolocC";
			String prefixB = nameB + ".";
			String nameC = "org.objectweb.medor.optim.jorm.rdb.Ccoloc";
			String prefixC = nameC + ".";
			RdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(null,
					new QualifiedTable[]{tableA, tableC, tableB}, "");
			QueryTreeField idA = leaf.addRdbField(
					prefixA + "a_name", PTypeSpace.LONG, "A_PK", tableA);
			QueryTreeField b1ToA = leaf.addRdbField(
					"BtoA_1_FK", PTypeSpace.LONG, "BtoA_1_FK", tableB);
			QueryTreeField cToB = leaf.addRdbField(
					"CtoB_FK", PTypeSpace.STRING, "CtoB_FK", tableC);
			QueryTreeField idB = leaf.addRdbField(
					prefixB + "b_name", PTypeSpace.STRING, "B_PK", tableB);
			QueryTreeField fcField = leaf.addRdbField(
					prefixC + "fc", PTypeSpace.STRING, "FC_COL", tableC);
			Expression eq1 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b1ToA)
			);
			Expression eq2 = new Equal(
					new BasicFieldOperand(idB), new BasicFieldOperand(cToB)
			);
			Expression expRdb = new And(eq1, eq2);
			leaf.setQueryFilter(expRdb);
			QueryNode expected = new SelectProject("");
			SinglePName spn = new SinglePName(
					new BasicFieldOperand(idA),
					new BasicParameterOperand(
							PTypeSpaceMedor.NAMING_CONTEXT, prefixA + "object"),
					ea.extent.getField(prefixA + "object").getType());
			expected.addCalculatedField(prefixA + "object", spn.getType(), spn);
			expected.addPropagatedField(prefixC + "fc", fcField.getType(), new QueryTreeField[]{fcField});

			logger.log(BasicLevel.DEBUG, "testAtoB: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
			logger.log(BasicLevel.DEBUG, "End of testAtoB");


		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}
	
	public void testAtoBtoCOrderBy() {
		try {
			Example a = getExtentOfAcolocB();
			Example b = getExtentOfBcolocC();
			Example c = getExtentOfCcoloc();

			Expression exp2 = new Equal(
					new BasicFieldOperand(b.extent.getField(b.clazz.getFQName() + ".c")),
					new BasicFieldOperand(c.extent.getField(c.extent.getPNameFieldName()))
			);
			QueryNode qn2 = new SelectProject(exp2, "");
			QueryTreeField bf1 = (QueryTreeField)b.extent.getField(b.clazz.getFQName() + ".f1");
			QueryTreeField bf1OnQn2 =
					qn2.addPropagatedField(bf1.getName(), bf1.getType(), new QueryTreeField[]{bf1});

			Expression exp1 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b1")),
					new BasicFieldOperand(bf1OnQn2)
			);
			QueryNode qn1 = new SelectProject(exp1, "");
			QueryTreeField f = (QueryTreeField) a.extent.getField(a.extent.getPNameFieldName());
			qn1.addPropagatedField(f.getName(), f.getType(), new QueryTreeField[]{f});
			QueryTreeField bcOnQn1 = qn1.addPropagatedField(bf1OnQn2.getName(), bf1OnQn2.getType(), new QueryTreeField[]{bf1OnQn2});


			//Add an additional level
			QueryNode qn = new SelectProject(exp1, "");
			QueryTreeField ff = (QueryTreeField)qn1.getTupleStructure().getField(a.extent.getPNameFieldName());
			qn.addPropagatedField(ff.getName(), ff.getType(), new QueryTreeField[]{ff});

			// Add an OrderBy Field
			OrderField of1 = new BasicOrderField(bcOnQn1, true);
			qn.setOrderBy(new OrderField[] {of1});

			QueryRewriter jqr = this.createQueryRewriter();
			logger.log(BasicLevel.DEBUG, "testBtoCOrderBy: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			a.extent.setPMapper(mapper, PROJECT_NAME);
			a.extent.setProjectName("test");
			b.extent.setPMapper(mapper, PROJECT_NAME);
			b.extent.setProjectName("test");
			c.extent.setPMapper(mapper, PROJECT_NAME);
			c.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			assertEquals("Bad sql request",
				"SELECT T_A_0.A_PK FROM T_A as T_A_0 left outer join T_B as T_B_1_b1_join on T_A_0.A_PK = T_B_1_b1_join.BtoA_1_FK, T_B as T_B_3 WHERE (((T_B_1_b1_join.B_PK singlePName ?) = T_B_3.col_f1) AND ((T_B_1_b1_join.B_PK singlePName ?) = T_B_3.col_f1)) ORDER BY T_B_3.col_f1 DESC",
				((RdbExpQueryLeaf)
					((QueryNode)result).getChildren()[0]).getSqlRequest(null,
							false, false));
			logger.log(BasicLevel.DEBUG, "End of testAtoBtoCOrderBy");


		} catch (Exception e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testCountAll() {
		try {
			Example e = getExtentOfAgg();
			QueryNode qn = new SelectProject("");
			QueryTreeField qtfpn = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			QueryTreeField fpn = qn.addPropagatedField(
					qtfpn.getName(),
					qtfpn.getType(),
					new QueryTreeField[]{qtfpn});

			QueryTreeField qtfy = (QueryTreeField)
			e.extent.getField(e.clazz.getFQName() + ".y");
			QueryTreeField fy = qn.addPropagatedField(
					qtfy.getName(),
					qtfy.getType(),
					new QueryTreeField[]{qtfy});

			Nest nest = new Nest(
					new QueryTreeField[] {fpn},
					"grouped",
					//new QueryTreeField[] {},
					new QueryTreeField[] {fy},
					"nestNode");
			//count with distinct
			nest.addCalculatedField("count",
					PTypeSpaceMedor.INT,
					new Count(
							new BasicFieldOperand(fpn), true
					));

			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testCountAll: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(nest, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			logger.log(BasicLevel.DEBUG, "testCountAll: Performing transformation");
			QueryTree result = jqr.transform(nest);
			//QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafAgg(e);
			QueryTreeField expC2 =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".c2");
			QueryTreeField expC1 =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".c1");
			QueryTreeField expY =
				(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".y");
			leaf.setDistinct(true);
			((RdbQueryLeaf)leaf).setIsSubquery(true);
			QueryNode expected = new AggregateRdbQueryNode(
					new QueryTreeField[]{expC1, expC2},
							"grouped",
							new QueryTreeField[] {expY},
							(BasicRdbExpQueryLeaf) leaf,
							null, "nestNode");
			Count theCount = new Count(
							new BasicFieldOperand(expC2), true
					);
			theCount.setCountAll();
			expected.addCalculatedField("count",
					PTypeSpaceMedor.INT,
					theCount
					);

			logger.log(BasicLevel.DEBUG, "testCountAll: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}
	
	public void testCountWithParam() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".object")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));

			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtfpn = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			QueryTreeField fpn = qn.addPropagatedField(
					qtfpn.getName(),
					qtfpn.getType(),
					new QueryTreeField[]{qtfpn});

			Nest nest = new Nest(
					new QueryTreeField[] {fpn},
					"grouped",
					new QueryTreeField[] {},
					"nestNode");
			//count with distinct
			nest.addCalculatedField("count",
					PTypeSpaceMedor.INT,
					new Count(
							new BasicFieldOperand(fpn), false
					));

			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testCountAll: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(nest, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(nest);
			//QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);

			QueryLeaf leaf = getQueryLeafAggParam(e);
			QueryTreeField qtfAid = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());

			QueryTreeField expId =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".id");
			((RdbQueryLeaf)leaf).setIsSubquery(true);

			((BasicRdbExpQueryLeaf)leaf).setQueryFilter(
					new Equal(
							new BasicFieldOperand(leaf.getTupleStructure().getField(e.clazz.getFQName() + ".id")),
							new EncodePName(PTypeSpace.LONG, "?1")
					)
			);
			QueryNode expected = new AggregateRdbQueryNode(
					new QueryTreeField[]{qtfAid},
							"grouped",
							new QueryTreeField[] {},
							(BasicRdbExpQueryLeaf) leaf,
							null, "nestNode");

			expected.addCalculatedField("count",
					PTypeSpaceMedor.INT,
					new Count(
							new BasicFieldOperand(expId), false
					)
					);

			logger.log(BasicLevel.DEBUG, "testCountAll: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testAvgDirect() {
		try {
			Example e = getExtentOfAgg();

			QueryTreeField qtf1 = (QueryTreeField)
					e.extent.getField(e.clazz.getFQName() + ".c1");

			QueryTreeField qtfy = (QueryTreeField)
				e.extent.getField(e.clazz.getFQName() + ".y");

			Nest nest = new Nest(
					new QueryTreeField[] {qtf1},
					"grouped",
					new QueryTreeField[] {qtfy},
					"nestNode");
			//count with distinct
			nest.addCalculatedField("avg",
					PTypeSpaceMedor.INT,
					new Avg(
							new BasicFieldOperand(qtf1), true
					));

			nest.addPropagatedField(qtfy.getName(), qtfy.getType(), new QueryTreeField[]{qtfy});
			
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testAvgDirect: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(nest, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(nest);
			//QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "testAvgDirect: QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafAggDirect(e);
			QueryTreeField expC1 =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".c1");
			QueryTreeField expY =
				(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".y");
			//((RdbQueryLeaf)leaf).setIsSubquery(false);
			QueryNode expected = new AggregateRdbQueryNode(
					new QueryTreeField[]{expC1},
							"grouped",
							new QueryTreeField[] {expY},
							(BasicRdbExpQueryLeaf) leaf,
							null, "nestNode");
			Avg theAvg = new Avg(
							new BasicFieldOperand(expC1), true
					);
			expected.addCalculatedField("avg",
					PTypeSpaceMedor.INT,
					theAvg
					);
			expected.addPropagatedField(expY.getName(), expY.getType(), new QueryTreeField[]{expY});


			logger.log(BasicLevel.DEBUG, "testAvgDirect: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}
	
	public void testAvgSum() {
		try {
			Example e = getExtentOfAgg();

			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".c2")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "");

			QueryTreeField qtf1 = (QueryTreeField)
					e.extent.getField(e.clazz.getFQName() + ".c1");
			QueryTreeField fc1 = qn.addPropagatedField(
					qtf1.getName(),
					qtf1.getType(),
					new QueryTreeField[]{qtf1});
			QueryTreeField qtfy = (QueryTreeField)
					e.extent.getField(e.clazz.getFQName() + ".y");
			QueryTreeField fcy = qn.addPropagatedField(
					qtfy.getName(),
					qtfy.getType(),
					new QueryTreeField[]{qtfy});

			QueryTreeField qtfpn = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());


			Nest nest = new Nest(
					new QueryTreeField[] {fc1, fcy},
					"grouped",
					new QueryTreeField[] {},
					"nestNode");
			nest.addCalculatedField("avg",
					PTypeSpaceMedor.FLOAT,
					new Avg(
							new BasicFieldOperand(fc1), true
					));
			nest.addCalculatedField("sum",
					PTypeSpaceMedor.FLOAT,
					new Sum(
							new BasicFieldOperand(fcy)
					));

			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testAvgSum: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(nest, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(nest);
			//QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "testAvgSum: QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);

			//testing indexes generation
			IndexesGenerator ig = new IndexesGenerator();
			ig.transform(result);

			QueryLeaf leaf = getQueryLeafAggWithWhere(e);

			QueryTreeField expC1 =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".c1");
			QueryTreeField expy =
					(QueryTreeField) leaf.getTupleStructure().getField(e.clazz.getFQName() + ".y");
			QueryNode expected = new AggregateRdbQueryNode(
					new QueryTreeField[]{expC1, expy},
							"grouped",
							new QueryTreeField[] {},
							(BasicRdbExpQueryLeaf) leaf,
							null, "nestNode");
			expected.addCalculatedField("avg",
					PTypeSpaceMedor.FLOAT,
					new Avg(
							new BasicFieldOperand(expC1), true
					));
			expected.addCalculatedField("sum",
					PTypeSpaceMedor.FLOAT,
					new Sum(
							new BasicFieldOperand(expy)
					));

			logger.log(BasicLevel.DEBUG, "testAvgSum: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void _testInheritance() {
		try {
			Example e = getExtentOfAnimal();
			e.extent.setWithSubClasses(true);
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".species")),
					new BasicOperand("marsupial"));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testInheritance: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			logger.log(BasicLevel.DEBUG, "End of QueryTree rewritten");
			QueryLeaf leaf = getQueryLeafAnimal(e);
			QueryNode expected = new SelectProject("");
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "animal_id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);

			logger.log(BasicLevel.DEBUG, "testInheritance: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}

	}

	public void _testInheritanceNF() {
		try {
			Example e = getExtentOfAnimalNF();
			e.extent.setWithSubClasses(true);
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".species")),
					new BasicOperand("marsupial"));
			QueryNode qn = new SelectProject(exp, "");
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryRewriter jqr = this.createQueryRewriter();

			logger.log(BasicLevel.DEBUG, "testInheritance: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			logger.log(BasicLevel.DEBUG, "End of QueryTree rewritten");
			QueryLeaf leaf = getQueryLeafAnimal(e);
			QueryNode expected = new SelectProject("");
			//TODO use composite pname instead
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "animal_id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);

			logger.log(BasicLevel.DEBUG, "testInheritance: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}

	}

}
